package securityutils;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.security.GeneralSecurityException;
import java.security.SecureRandom;
import java.util.zip.Deflater;
import java.util.zip.Inflater;
import java.util.zip.DataFormatException;

import javax.crypto.Cipher;
import javax.crypto.Mac;
import javax.crypto.SecretKey;
import javax.crypto.spec.IvParameterSpec;

import org.bouncycastle.util.encoders.Base64Encoder;

/**
 * A class for safely storing data, even when untrustworthy entities can
 * get access to the data.
 * 
 * <h2>Introduction</h2>
 *
 * <p>This class provides a simple way to securely protect data from
 * eavesdropping and tampering. Most methods and fields are private;
 * programmers using this class should use the public methods only. The
 * public fields are essentially informational and not hugely relevant to
 * the application programmer.</p>
 *
 * <p>{@link #seal} provides <em>confidentiality</em> and
 * <em>integrity</em>.  To read the sealed data, a malicious user would need
 * the encryption key. To mangle it or carefully change it (without being
 * detected by {@link #unseal}), they would need the signing key.
 * Additionally, <code>unseal</code> optionally ensures freshness: it will
 * reject data that was sealed more than <code>expirationWindow</code>
 * milliseconds before.</p>
 *
 * <p><strong>PLEASE NOTE!</strong> This class does not, by itself, protect
 * against <em>replay attacks</em> &mdash; attacks in which the attacker
 * replays a valid sealed blob within the expiration window. You can
 * mitigate <strong>but not completely prevent</strong> this attack by
 * setting a very low expiration window. You can fully prevent it by putting
 * application-specific data, such as a one-time transaction ID, in the
 * sealed blob. Then, your business logic code can detect the replayed
 * transaction. But <code>Seal</code> by itself cannot do this.
 *
 * <p>To use the class, pass your data to <code>seal</code> to get a sealed
 * <em>blob</em>. The blob is a printable (base64-encoded) String of binary
 * data. Then you can reconstitute the blob by passing it to
 * <code>unseal</code>.</p>
 *
 * <p>You may also want to compress the data, in case it is large and you
 * need to fit it into a small space (like an HTTP cookie). For this
 * purpose, we provide the convenience methods {@link #compress} and {@link
 * #decompress}. <strong>NOTE:</strong> You must compress before sealing,
 * and decompress after unsealing! Encrypted data will not compress!</p>
 *
 * <h2>Applications</h2>
 *
 * <p>This class is useful when you want to take trusted state &mdash; that
 * is, state you depend on to be correct for your application to work safely
 * &mdash; out of a trustworthy system. For example, web applications often
 * require many hits on the user session state table in the database, and
 * this can be a performance bottleneck. As a result, web developers would
 * like to put the state itself in the cookie (rather than having the cookie
 * value be an index into the state table, as with normal
 * <em>JSESSIONID</em>-type cookies). However, this is not by itself safe!
 * The user can modify their cookie and thereby modify what the server
 * application thinks the user's session state is. The user can also read
 * the session state, which may contain sensitive server information. We
 * have seen and performed this attack many times in our work.</p>
 *
 * <p>However, this class makes that practice much safer. The encryption and
 * integrity checking ensure that, when the server unseals the data, the
 * user (or whatever untrustworthy entity, like another computer or a
 * network attacker) cannot have decrypted the sealed blob. Furthermore, if
 * they modified the blob, that modification will be detected during the
 * unsealing process, and <code>unseal</code> will throw a
 * <code>CorruptedBlobException</code>.</p>
 *
 * <p>Still, keep in mind the replay attack. Make sure that replays won't
 * hurt your application, mitigate them by setting an appropriate expiration
 * window, and/or use one-time transaction IDs for important transactions.</p>
 * 
 * @author Chris Palmer <chris@isecpartners.com>
 * @version 1
 */
abstract public class Seal {

      /** The block cipher used to protect the confidentiality of sealed
       * data. */
      public static final String BLOCK_CIPHER = "AES/CBC/PKCS7Padding";

      /** The HMAC used to protect the integrity of sealed data. */
      public static final String HMAC_ALGORITHM = "HMACSHA256";

      private static final int BLOCK_SIZE = 16;
      private static final int KEY_SIZE = 16;
      private static final double BASE64_EXPANSION_RATIO = 1.37;
      private static final String FIELD_SEPARATOR = "|";
      private static final SecureRandom SECURE_RANDOM = new SecureRandom();


      /**
       * @param data The data to compress.
       *
       * @return The compressed data.
       *
       * @throws IOException
       */
      public static byte [] compress(byte [] data) throws IOException {
            Deflater dfltr = new Deflater(Deflater.BEST_COMPRESSION);
            dfltr.setInput(data);
            dfltr.finish();
            ByteArrayOutputStream outpt = new ByteArrayOutputStream(data.length);

            byte [] bfr = new byte[1024];
            while (! dfltr.finished()) {
                  int c = dfltr.deflate(bfr);
                  outpt.write(bfr, 0, c);
            }
            outpt.close();

            return outpt.toByteArray();
      }


      /**
       * @param data The data to decompress.
       *
       * @return The decompressed data.
       *
       * @throws IOException
       * @throws DataFormatException
       */
      public static byte [] decompress(byte [] data) throws IOException, DataFormatException {
            Inflater infltr = new Inflater();
            infltr.setInput(data);
            ByteArrayOutputStream outpt = new ByteArrayOutputStream(data.length);

            byte [] bfr = new byte[1024];
            while (! infltr.finished()) {
                  int c = infltr.inflate(bfr);
                  outpt.write(bfr, 0, c);
            }
            outpt.close();

            return outpt.toByteArray();
      }


      /**
       * @param mode One of Cipher.ENCRYPT_MODE or Cipher.DECRYPT_MODE.
       * @param data The plaintext to be encrypted or the ciphertext to be
       * decrypted.
       * @param key The encryption/decryption key.
       * @param iv A strongly random (as from SecureRandom) initialization
       * vector.
       *
       * @return The encrypted/decrypted data.
       *
       * @throws GeneralSecurityException
       */
      private static byte [] encrypt(int mode, byte [] data, SecretKey key, byte [] iv)
            throws GeneralSecurityException
      {
            Cipher cphr = Cipher.getInstance(BLOCK_CIPHER);

            int bs = cphr.getBlockSize();
            if (iv.length < bs)
                  throw new Error("iv.length was " + iv.length + " bytes, needed " +
                              bs);

            cphr.init(mode, key, new IvParameterSpec(iv, 0, bs), SECURE_RANDOM);
            return cphr.doFinal(data);
      }


      /**
       * @param data The data to HMAC.
       * @param key The secret signing key.
       *
       * @return The HMAC of the data.
       *
       * @throws GeneralSecurityException
       */
      private static byte [] sign(byte [] data, byte [] timestamp, SecretKey key)
            throws GeneralSecurityException
      {
            Mac mc = null;
            mc = Mac.getInstance(HMAC_ALGORITHM);

            mc.init(key);
            mc.update(timestamp);
            return mc.doFinal(data);
      }


      /**
       * @param data The data whose signature you want to check.
       * @param key The secret signing key.
       * @param signature The purported data signature.
       *
       * @return true if the signature is the correct HMAC of the data, given
       * key.
       *
       * @throws GeneralSecurityException
       */
      private static boolean signatureValid(byte [] data, byte [] timestamp, SecretKey key,
                                            byte [] signature)
            throws GeneralSecurityException
      {

            /* We can't use this because it creates a timing vulnerability:
             * An attacker can guess where their guessed HMAC is wrong
             * by seeing how quickly this function returns.
             * Therefore, we have to check every byte, even after we've
             * found a mismatch.
             *
             * See Nate Lawson's blog post at
             * http://rdist.root.org/2009/05/28/timing-attack-in-google-keyczar-library/
             * for more information.
             *
            return Arrays.equals(signature, sign(data, timestamp, key)); */

            byte [] trstd = sign(data, timestamp, key);
            int ln = trstd.length;
            if (signature.length != ln)
                  return false;

            int r = 0;
            for (int i = 0; i < ln; i++)
                  r |= signature[i] ^ trstd[i];

            return 0 == r;
      }


      /**
       * @return The current time in milliseconds since the epoch, in String
       * representation (base 10).
       */
      private static String timestamp() {
            return Long.toString(System.currentTimeMillis());
      }


      /**
       * @param data The data to encode.
       *
       * @return The encoded data.
       *
       * @throws IOException
       */
      protected static String base64Encode(byte [] data) throws IOException {
            Base64Encoder encdr = new Base64Encoder();
            ByteArrayOutputStream outpt = new ByteArrayOutputStream(
                        (int) Math.ceil(data.length * BASE64_EXPANSION_RATIO));
            encdr.encode(data, 0, data.length, outpt);

            return outpt.toString();
      }


      /**
       * @param data The base64-encoded data to decode.
       *
       * @return The decoded data.
       *
       * @throws CorruptBlobException
       */
      protected static byte [] base64Decode(String data) throws CorruptBlobException {
            try {
                  Base64Encoder encdr = new Base64Encoder();
                  ByteArrayOutputStream outpt = new ByteArrayOutputStream(
                              (int) Math.ceil(data.length() / BASE64_EXPANSION_RATIO));
                  encdr.decode(data, outpt);

                  return outpt.toByteArray();
            }
            catch (Throwable t) {
                  throw new CorruptBlobException(t);
            }
      }


      /**
       * @param data The plaintext data to encrypt, timestamp, and integrity
       * check.
       * @param encryptionKey The key to encrypt data.
       * @param signingKey The key to HMAC data.
       *
       * @return A base64-encoded representation of the sealed data.
       * 
       * @throws IOException
       * @throws GeneralSecurityException
       */
      public static String seal(byte [] data, SecretKey encryptionKey, SecretKey signingKey)
            throws IOException, GeneralSecurityException
      {
            byte [] iv = new byte[BLOCK_SIZE];
            SECURE_RANDOM.nextBytes(iv);

            byte [] ncrptd = encrypt(Cipher.ENCRYPT_MODE, data, encryptionKey, iv);

            byte [] cphrtxt = new byte[BLOCK_SIZE + ncrptd.length];
            System.arraycopy(iv, 0, cphrtxt, 0, BLOCK_SIZE);
            System.arraycopy(ncrptd, 0, cphrtxt, BLOCK_SIZE, ncrptd.length);

            String tmstmp = timestamp();
            byte [] sgntr = sign(cphrtxt, tmstmp.getBytes(), signingKey);

            return base64Encode(sgntr) + FIELD_SEPARATOR +
                        tmstmp + FIELD_SEPARATOR +
                        base64Encode(cphrtxt);
      }


      /**
       * @param data A sealed data blob created with seal.
       * @param expirationWindow The validity period of data in milliseconds.  If
       * 0, data never expires.
       * @param encryptionKey The key that was used to encrypt data.
       * @param signingKey The key that was used to sign data.
       *
       * @return The decrypted data.
       *
       * @throws CorruptBlobException if there is anything wrong with data:
       * integrity check failure, timestamp expired, decryption failure, or
       * anything else.
       *
       */
      public static byte [] unseal(String data, long expirationWindow,
            SecretKey encryptionKey, SecretKey signingKey) throws CorruptBlobException
      {
            int mcEnd = data.indexOf(FIELD_SEPARATOR);
            int tmstmpEnd = data.lastIndexOf(FIELD_SEPARATOR);
            if (-1 == mcEnd || -1 == tmstmpEnd || mcEnd == tmstmpEnd)
                  throw new CorruptBlobException("Invalid data format");

            try {
                  byte [] mc = base64Decode(data.substring(0, mcEnd));
                  String tmstmp = data.substring(mcEnd + 1, tmstmpEnd);
                  byte [] cphrtxt = base64Decode(data.substring(tmstmpEnd + 1));

                  if (! signatureValid(cphrtxt, tmstmp.getBytes(), signingKey, mc))
                        throw new CorruptBlobException("failed integrity check");

                  // As we approach Sat Aug 16 23:12:55 PST 292278994, this code
                  // will start to have problems.
                  if (0 != expirationWindow) {
                        long tm = Long.parseLong(tmstmp);
                        long lmt = tm + expirationWindow;
                        if (tm < 0 || lmt <= tm || System.currentTimeMillis() > lmt)
                              throw new CorruptBlobException("data expired");
                  }

                  byte [] iv = new byte [BLOCK_SIZE];
                  byte [] ncrptd = new byte [cphrtxt.length - BLOCK_SIZE];
                  System.arraycopy(cphrtxt, 0, iv, 0, BLOCK_SIZE);
                  System.arraycopy(cphrtxt, BLOCK_SIZE, ncrptd, 0,
                        cphrtxt.length - BLOCK_SIZE);

                  return encrypt(Cipher.DECRYPT_MODE, ncrptd, encryptionKey, iv);
            }
            catch (GeneralSecurityException e) {
                  throw new CorruptBlobException(e);
            }
      }
}

