package securityutils;

import java.io.File;
import java.io.FileInputStream;
import java.io.IOException;
import java.security.NoSuchAlgorithmException;
import java.security.NoSuchProviderException;
import java.security.Security;
import java.util.Random;

import javax.crypto.SecretKey;
import javax.crypto.KeyGenerator;

import junit.framework.*;
import org.bouncycastle.jce.provider.BouncyCastleProvider;


public class SealTest extends TestCase {

      static Random r = new Random();
      static String randomString() {
            int c = r.nextInt() % 256;
            StringBuilder s = new StringBuilder();

            for (int i = 0; i < c; i++)
                  s.append(Long.toString(Math.abs(r.nextLong()), 64));

            return s.toString();
      }

      static {
            Security.addProvider(new BouncyCastleProvider());
      }

      public SecretKey signingKey;
      public SecretKey encryptionKey;
      public static final long EXPIRATION_WINDOW = 1500;
      public static final String testData = "This is a simple test string";
      

      public SealTest() {
            super();
      }
      

      public SealTest(String test) {
            super(test);
      }
      

      protected void setUp()
            throws NoSuchAlgorithmException, NoSuchProviderException, IOException
      {
            KeyGenerator kgen = KeyGenerator.getInstance("AES", "BC");
            kgen.init(128);
            encryptionKey = kgen.generateKey();
            signingKey = KeyGenerator.getInstance(Seal.HMAC_ALGORITHM, "BC").generateKey();
            System.out.println( "encryptionKey: " +
                  Seal.base64Encode(encryptionKey.getEncoded()) +
                  " (" + encryptionKey.getAlgorithm() + ", " +
                  encryptionKey.getFormat() + ")" );
            System.out.println( "signingKey: " +
                  Seal.base64Encode(signingKey.getEncoded()) +
                  " (" + signingKey.getAlgorithm() + ", " +
                  signingKey.getFormat() + ")" );
      }


      public void testCompressOnly(String testData) {
            try {
                  byte [] cmprsd = Seal.compress(testData.getBytes("UTF-8"));
                  String s = Seal.base64Encode(cmprsd);
            } catch (Exception e) {
                  fail("Unexpected exception: " + e);
            }
      }


      public void testNormal(byte [] s) {
            try {
                  String sld = Seal.seal(s, encryptionKey, signingKey);
                  String unsld = new String(Seal.unseal(sld, EXPIRATION_WINDOW, encryptionKey,
                                                        signingKey));
                  assertTrue(unsld.equals(new String(s, "UTF-8")));
            } catch (Exception e) {
                  fail("Unexpected exception: " + e);
            }
      }


      public void testNormal(String s) {
            try {
                  testNormal(s.getBytes("UTF-8"));
            } catch (Exception e) {
                  fail("Unexpected exception: " + e);
            }
      }


      public void testExpired() {
            try {
                  String sld = Seal.seal(testData.getBytes("UTF-8"), encryptionKey, signingKey);
                  Thread.sleep(EXPIRATION_WINDOW + 10);
                  String unsld = new String(Seal.unseal(sld, EXPIRATION_WINDOW, encryptionKey,
                                                        signingKey));

                  fail("Successfully unsealed expired data!: " + unsld);
            } catch (CorruptBlobException e) {
                  // Expected.
            } catch (Exception e) {
                  fail("Unexpected exception: " + e);
            }
      }


      public void testJunk() {
            try {
                  String sld = Seal.seal(testData.getBytes("UTF-8"), encryptionKey, signingKey);
                  sld += "This is junk to ruin the HMAC.";
                  String unsld = new String(Seal.unseal(sld, EXPIRATION_WINDOW, encryptionKey,
                                                        signingKey));
                  fail("Successfully unsealed junk data!: " + unsld);
            } catch (CorruptBlobException e) {
                  // Expected.
            } catch (Exception e) {
                  fail("Unexpected exception: " + e);
            }
      }


      public void testJunk(String s) {
            try {
                  String sld = Seal.seal(s.getBytes("UTF-8"), encryptionKey, signingKey);
                  sld += "This is junk to ruin the HMAC.";
                  String unsld = new String(Seal.unseal(sld, EXPIRATION_WINDOW, encryptionKey,
                                                        signingKey));
                  fail("Successfully unsealed junk data!: " + unsld);
            } catch (CorruptBlobException e) {
                  // Expected.
            } catch (Exception e) {
                  fail("Unexpected exception: " + e);
            }
      }


      public void testUnicodeJunk() {
            try {
                  String sld = Seal.seal(testData.getBytes("UTF-8"), encryptionKey, signingKey);
                  sld += "This is multi-byte character junk to confuse the Base64 size calculations: \u1234 \u5353 \u4321";
                  String unsld = new String(Seal.unseal(sld, EXPIRATION_WINDOW, encryptionKey,
                                                        signingKey));
                  fail("Successfully unsealed junk data!: " + unsld);
            } catch (CorruptBlobException e) {
                  // Expected.
            } catch (Exception e) {
                  fail("Unexpected exception: " + e);
            }
      }


      public static byte [] readFile(File file) throws Exception {
            int lngth = (int) file.length();
            byte [] data = new byte [lngth];
            FileInputStream fis = new FileInputStream(file);
            fis.read(data);
            fis.close();
            return data;
      }


      public static void main(String [] args) {
            final int reps = 2000;
            try {
                  SealTest t = new SealTest();
                  t.setUp();
                  t.testNormal(testData);
                  t.testJunk();
                  t.testUnicodeJunk();
                  t.testExpired();

                  System.out.println("Tests passed! Now to demonstrate performance. (" + reps +
                                     " iterations)");

                  long tm;

                  tm = System.currentTimeMillis();
                  for (int i = 0; i < reps; i++)
                        t.testNormal(testData);
                  System.out.println("Time (testNormal(testData(testData)): " + ( System.currentTimeMillis() - tm) +
                                     " ms");

                  tm = System.currentTimeMillis();
                  for (int i = 0; i < reps; i++)
                        t.testCompressOnly(testData);
                  System.out.println("Time (testCompressOnly(testData)): " +
                                     (System.currentTimeMillis() - tm) +
                                     " ms");

                  tm = System.currentTimeMillis();
                  for (int i = 0; i < reps; i++)
                        t.testCompressOnly(randomString());
                  System.out.println("Time (testCompressOnly(randomString)): " +
                                     (System.currentTimeMillis() - tm) +
                                     " ms");

                  tm = System.currentTimeMillis();
                  for (int i = 0; i < reps; i++)
                        t.testJunk();
                  System.out.println("Time (testJunk): " + ( System.currentTimeMillis() - tm) +
                                     " ms");

                  tm = System.currentTimeMillis();
                  for (int i = 0; i < reps; i++)
                        t.testNormal(randomString());
                  System.out.println("Time (testNormal(randomString)): " +
                                     ( System.currentTimeMillis() - tm) + " ms");

                  tm = System.currentTimeMillis();
                  for (int i = 0; i < reps; i++)
                        t.testJunk(randomString());
                  System.out.println("Time (testJunk(randomString)): " +
                                     ( System.currentTimeMillis() - tm) + " ms");


                  System.out.println("Reading file " + args[0]);
                  String flData = new String(readFile(new File(args[0])), "UTF-8");
                  System.out.println("File contains " + flData.length() + " characters");

                  tm = System.currentTimeMillis();
                  for (int i = 0; i < reps; i++)
                        t.testNormal(flData);
                  System.out.println("Time (testNormal(fileData)): " +
                                     ( System.currentTimeMillis() - tm) + " ms");

                  tm = System.currentTimeMillis();
                  for (int i = 0; i < reps; i++)
                        t.testCompressOnly(flData);
                  System.out.println("Time (testCompressOnly(fileData)): " +
                                     ( System.currentTimeMillis() - tm) + " ms");
            }
            catch (Exception e) {
                  System.err.println(e.toString());
                  System.exit(1);
            }
      }

}

