#!/usr/bin/env python

"""
An HTML parser that validates that the input adheres to a limited subset of
HTML.

Command line usage:

      HTMLValidator.py input.html > filtered.html
"""


from HTMLParser import HTMLParser


class HTMLSecurityViolation(Exception):
      pass


class HTMLValidator(HTMLParser):

      """A security-validating HTML parser."""

      def __init__(self, policy, tolerant):

            """policy: A dictionary of element-name: (attribute-names)
describing the approved HTML subset. tolerant: Whether or not to raise an
exception when invalid input is seen, or to simply ignore it and filter it
from the input."""

            HTMLParser.__init__(self)
            self.policy = policy
            self.tolerant = tolerant
            self.output = ""
            self.bad_tag = None

      def handle_starttag(self, tag, attributes):
            pstn = self.getpos()
            if tag not in self.policy:
                  if not self.tolerant:
                        raise HTMLSecurityViolation("Illegal tag \"%s\" occured at line %d, character %d"
                                                    % (tag, pstn[0], pstn[1]))
                  else:
                        # XXX: We don't handle nesting. :(
                        self.bad_tag = tag
                        return

            for a in attributes:
                  if a[0] not in self.policy[tag]:
                        if not self.tolerant:
                              raise HTMLSecurityViolation(
                                  "Illegal attribute \"%s\" occured in tag \"%s\" at line %d, character %d"
                                  % (a[0], tag, pstn[0], pstn[1]))
                        else:
                              self.bad_tag = tag
                              return

            self.output += self.get_starttag_text()

      def handle_startendtag(self, tag, attributes):
            self.handle_starttag(tag, attributes)
            # Don't call self.handle_endtag. We are done.

      def handle_endtag(self, tag):
            if tag == self.bad_tag:
                  return

            pstn = self.getpos()
            if tag not in self.policy:
                  if not self.tolerant:
                        raise HTMLSecurityViolation("Illegal tag \"%s\" occured at line %d, character %d"
                                                    % (tag, pstn[0], pstn[1]))
                  else:
                        return

            self.output += "</" + tag + ">"

      def handle_data(self, data):
            self.output += data

      def handle_charref(self, name):
            self.output += "&#" + name + ";"

      def handle_entityref(self, name):
            self.output += "&" + name + ";"

      def handle_comment(self, data):

            """We ignore comments. This is safest, because comments are one
way to hide code, yet browsers might try to execute it anyway."""

            return

      def handle_decl(self, data):

            """If tolerant, ignore declarations. Otherwise, raise
HTMLSecurityViolation."""

            if self.tolerant:
                  return
            raise HTMLSecurityViolation("Illegal SGML declaration at line %d, character %d"
                                        % self.getpos())

      def handle_pi(self, data):

            """If tolerant, ignore processing instructions. Otherwise, raise
HTMLSecurityViolation."""

            if self.tolerant:
                  return
            raise HTMLSecurityViolation("Illegal processing instruction at line %d, character %d"
                                        % self.getpos())


if __name__ == "__main__":

      from sys import argv as arguments, exit

      if 2 != len(arguments):
            print __doc__
            exit(1)

      p = HTMLValidator({ "p": ("class"), "b": (), "em": (), "hr": ("size"),
                          "h1": ("class"), "h2": ("class"), "h3": ("class"),
                          "h4": ("class"), "h5": ("class"),
                          "strong": (), "i": (), "ul": (), "ol": (),
                          "dl": (), "li": ("class"), "dd": ("class"),
                          "dt": ("class"), "br": (), "pre": () },
                         True)

      p.feed(file(arguments[1], "rb").read())
      p.close()
      print p.output

